# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import Counter
import math
import random
import warnings

import numpy as np

from ..gate import *
from ..utils import ignore_global_phase
from .backendbase import Backend

DEFAULT_DTYPE = complex


class _NumPyBackendContext:
    """This class is internally used in NumPyBackend"""

    def __init__(self, n_qubits):
        self.n_qubits = n_qubits
        self.qubits = np.zeros(2**n_qubits, dtype=DEFAULT_DTYPE)
        self.qubits_buf = np.zeros(2**n_qubits, dtype=DEFAULT_DTYPE)
        self.indices = np.arange(2**n_qubits, dtype=np.uint32)
        self.save_cache = True
        self.shots_result = Counter()
        self.cregs = None

    def prepare(self, cache):
        """Prepare to run next shot."""
        if cache is not None:
            np.copyto(self.qubits, cache)
        else:
            self.qubits.fill(0.0)
            self.qubits[0] = 1.0
        self.cregs = [0] * self.n_qubits

    def store_shot(self):
        """Store current cregs to shots_result"""
        def to_str(cregs):
            return ''.join(str(b) for b in cregs)
        key = to_str(self.cregs)
        self.shots_result[key] = self.shots_result.get(key, 0) + 1


class NumPyBackend(Backend):
    """Simulator backend which uses numpy. This backend is Blueqat's default backend."""
    __return_type = {
        "statevector": lambda ctx: ctx.qubits,
        "shots": lambda ctx: ctx.shots_result,
        "statevector_and_shots": lambda ctx: (ctx.qubits, ctx.shots_result),
        "_inner_ctx": lambda ctx: ctx,
    }
    DEFAULT_SHOTS = 1024

    def __init__(self):
        self.cache = None
        self.cache_idx = -1

    def __clear_cache(self):
        self.cache = None
        self.cache_idx = -1

    def __clear_cache_if_invalid(self, n_qubits, dtype):
        if self.cache is None:
            self.__clear_cache()
            return
        if len(self.cache) != 2**n_qubits:
            self.__clear_cache()
            return
        if self.cache.dtype != dtype:
            self.__clear_cache()
            return

    def run(self, gates, n_qubits, *args, **kwargs):
        def __parse_run_args(shots=None, returns=None, ignore_global=False, **_kwargs):
            if returns is None:
                if shots is None:
                    returns = "statevector"
                else:
                    returns = "shots"
            if returns not in self.__return_type.keys():
                raise ValueError(f"Unknown returns type '{returns}'")
            if shots is None:
                if returns in ("statevector", "_inner_ctx"):
                    shots = 1
                else:
                    shots = self.DEFAULT_SHOTS
            if returns == "statevector" and shots > 1:
                warnings.warn("When `returns` = 'statevector', `shots` = 1 is enough.")
            return shots, returns, ignore_global

        shots, returns, ignore_global = __parse_run_args(*args, **kwargs)

        self.__clear_cache_if_invalid(n_qubits, DEFAULT_DTYPE)
        ctx = _NumPyBackendContext(n_qubits)

        def run_single_gate(gate):
            nonlocal ctx
            action = self._get_action(gate)
            if action is not None:
                ctx = action(gate, ctx)
            else:
                for g in gate.fallback(n_qubits):
                    run_single_gate(g)

        for _ in range(shots):
            ctx.prepare(self.cache)
            for gate in gates[self.cache_idx + 1:]:
                run_single_gate(gate)
                if ctx.save_cache:
                    self.cache = ctx.qubits.copy()
                    self.cache_idx += 1
            if ctx.cregs:
                ctx.store_shot()

        if ignore_global:
            ignore_global_phase(ctx.qubits)
        return self.__return_type[returns](ctx)

    def make_cache(self, gates, n_qubits):
        self.run(gates, n_qubits)

    def gate_x(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = qubits[t1]
            newq[t1] = qubits[t0]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_y(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = -1.0j * qubits[t1]
            newq[t1] = 1.0j * qubits[t0]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_z(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= -1
        return ctx

    def gate_h(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = qubits[t0] + qubits[t1]
            newq[t1] = qubits[t0] - qubits[t1]
            newq *= 1 / np.sqrt(2)
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_rx(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = np.cos(halftheta)
        a01 = a10 = -1j * np.sin(halftheta)
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = a00 * qubits[t0] + a01 * qubits[t1]
            newq[t1] = a10 * qubits[t0] + a11 * qubits[t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_ry(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = np.cos(halftheta)
        a10 = np.sin(halftheta)
        a01 = -a10
        for target in gate.target_iter(n_qubits):
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = a00 * qubits[t0] + a01 * qubits[t1]
            newq[t1] = a10 * qubits[t0] + a11 * qubits[t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_rz(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a0 = complex(math.cos(halftheta), -math.sin(halftheta))
        a1 = complex(math.cos(halftheta), math.sin(halftheta))
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) == 0] *= a0
            qubits[(i & (1 << target)) != 0] *= a1
        return ctx

    def gate_phase(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        theta = gate.theta
        a = complex(math.cos(theta), math.sin(theta))
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= a
        return ctx

    def gate_t(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices

        sqrt2_inv = 1 / np.sqrt(2)
        factor = complex(sqrt2_inv, sqrt2_inv)
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= factor
        return ctx

    def gate_s(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) != 0] *= 1.j
        return ctx

    def gate_cz(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for control, target in gate.control_target_iter(n_qubits):
            qubits[((i & (1 << control)) != 0) & ((i & (1 << target)) != 0)] *= -1
        return ctx

    def gate_cx(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for control, target in gate.control_target_iter(n_qubits):
            np.copyto(newq, qubits)
            c1 = (i & (1 << control)) != 0
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[c1 & t0] = qubits[c1 & t1]
            newq[c1 & t1] = qubits[c1 & t0]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_crx(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = np.cos(halftheta)
        a01 = a10 = -1j * np.sin(halftheta)
        for control, target in gate.control_target_iter(n_qubits):
            np.copyto(newq, qubits)
            c1 = (i & (1 << control)) != 0
            c1t0 = ((i & (1 << target)) == 0) & c1
            c1t1 = ((i & (1 << target)) != 0) & c1
            newq[c1t0] = a00 * qubits[c1t0] + a01 * qubits[c1t1]
            newq[c1t1] = a10 * qubits[c1t0] + a11 * qubits[c1t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_cry(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a00 = a11 = np.cos(halftheta)
        a10 = np.sin(halftheta)
        a01 = -a10
        for control, target in gate.control_target_iter(n_qubits):
            np.copyto(newq, qubits)
            c1 = (i & (1 << control)) != 0
            c1t0 = ((i & (1 << target)) == 0) & c1
            c1t1 = ((i & (1 << target)) != 0) & c1
            newq[c1t0] = a00 * qubits[c1t0] + a01 * qubits[c1t1]
            newq[c1t1] = a10 * qubits[c1t0] + a11 * qubits[c1t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_crz(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halftheta = gate.theta * 0.5
        a0 = complex(math.cos(halftheta), -math.sin(halftheta))
        a1 = complex(math.cos(halftheta), math.sin(halftheta))
        for control, target in gate.control_target_iter(n_qubits):
            c1t0 = ((i & (1 << control)) != 0) & ((i & (1 << target)) == 0)
            c1t1 = ((i & (1 << control)) != 0) & ((i & (1 << target)) != 0)
            qubits[c1t0] *= a0
            qubits[c1t1] *= a1
        return ctx

    def gate_cphase(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        theta = gate.theta
        a = complex(math.cos(theta), math.sin(theta))
        for control, target in gate.control_target_iter(n_qubits):
            c1t1 = ((i & (1 << control)) != 0) & ((i & (1 << target)) != 0)
            qubits[c1t1] *= a
        return ctx

    def gate_ccz(self, gate, ctx):
        c1, c2, t = gate.targets
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        indices = (i & (1 << c1)) != 0
        indices &= (i & (1 << c2)) != 0
        indices &= (i & (1 << t)) != 0
        qubits[indices] *= -1
        return ctx

    def gate_ccx(self, gate, ctx):
        c1, c2, t = gate.targets
        ctx = self.gate_h(HGate(t), ctx)
        ctx = self.gate_ccz(CCZGate(gate.targets), ctx)
        ctx = self.gate_h(HGate(t), ctx)
        return ctx

    def gate_u1(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        halflambda = gate.lambd * 0.5
        a0 = complex(math.cos(halflambda), -math.sin(halflambda))
        a1 = complex(math.cos(halflambda), math.sin(halflambda))
        for target in gate.target_iter(n_qubits):
            qubits[(i & (1 << target)) == 0] *= a0
            qubits[(i & (1 << target)) != 0] *= a1
        return ctx

    def gate_u3(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        theta = gate.theta
        phi = gate.phi
        lambd = gate.lambd
        globalphase = complex(math.cos((-phi - lambd) * 0.5), math.sin((-phi - lambd) * 0.5))
        a00 = math.cos(theta * 0.5) * globalphase
        a11 = a00 * complex(math.cos(phi + lambd), math.sin(phi + lambd))
        a01 = a10 = math.sin(theta * 0.5) * globalphase
        a01 *= complex(math.cos(lambd), math.sin(lambd))
        a10 *= complex(math.cos(phi), math.sin(phi))
        for target in gate.target_iter(n_qubits):
            np.copyto(newq, qubits)
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = qubits[t0] * a00
            newq[t0] -= qubits[t1] * a01
            newq[t1] = qubits[t0] * a10
            newq[t1] += qubits[t1] * a11
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_mat1(self, gate, ctx):
        qubits = ctx.qubits
        newq = ctx.qubits_buf
        n_qubits = ctx.n_qubits
        i = ctx.indices
        mat = gate.mat
        for target in gate.target_iter(n_qubits):
            np.copyto(newq, qubits)
            t0 = (i & (1 << target)) == 0
            t1 = (i & (1 << target)) != 0
            newq[t0] = mat[0, 0] * qubits[t0]
            newq[t0] += mat[0, 1] * qubits[t1]
            newq[t1] = mat[1, 0] * qubits[t0]
            newq[t1] += mat[1, 1] * qubits[t1]
            qubits, newq = newq, qubits
        ctx.qubits = qubits
        ctx.qubits_buf = newq
        return ctx

    def gate_measure(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            p_zero = np.linalg.norm(qubits[(i & (1 << target)) == 0]) ** 2
            rand = random.random()
            if rand < p_zero:
                qubits[(i & (1 << target)) != 0] = 0.0
                qubits /= np.sqrt(p_zero)
                ctx.cregs[target] = 0
            else:
                qubits[(i & (1 << target)) == 0] = 0.0
                qubits /= np.sqrt(1.0 - p_zero)
                ctx.cregs[target] = 1
        ctx.save_cache = False
        return ctx

    def gate_reset(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        i = ctx.indices
        for target in gate.target_iter(n_qubits):
            p_zero = np.linalg.norm(qubits[(i & (1 << target)) == 0]) ** 2
            rand = random.random()
            t1 = (i & (1 << target)) != 0
            if rand < p_zero:
                qubits[t1] = 0.0
                qubits /= np.sqrt(p_zero)
            else:
                qubits[(i & (1 << target)) == 0] = qubits[t1]
                qubits[t1] = 0.0
                qubits /= np.sqrt(1.0 - p_zero)
        ctx.save_cache = False
        return ctx



from collections import Counter
from .backendbase import Backend
import blueqat.gate as bqgate
import numpy as np

class QgateBackend(Backend) :

    def __init__(self) :
        import sys
        this = sys.modules[__name__]
        if not hasattr(this, 'qgate') :
            import qgate
            this.qgate = qgate
            this.model = qgate.model
            this.gtype = qgate.model.gate_type

            QgateBackend.gatetypes = {
                'i' : (gtype.ID, 0, 1),
                'x' : (gtype.X, 0, 1), 'y' : (gtype.Y, 0, 1), 'z' : (gtype.Z, 0, 1),
                'h' : (gtype.H, 0, 1), 't' : (gtype.T, 0, 1), 's' : (gtype.S, 0, 1),
                # adjoint
                'tdg' : (gtype.T, 0, 1), 'sdg' : (gtype.S, 0, 1),
                # 1-control-bit gate
                'cz' : (gtype.Z, 1, 1), 'cx' : (gtype.X, 1, 1), 'cnot' : (gtype.X, 1, 1),
                # rotation, phase gate
                'rx' : (gtype.RX, 0, 1), 'ry' : (gtype.RY, 0, 1), 'rz' : (gtype.RZ, 0, 1),
                'phase' : (gtype.U1, 0, 1),
                # controlled rotation gate
                'crx' : (gtype.RX, 1, 1), 'cry' : (gtype.RY, 1, 1), 'crz' : (gtype.RZ, 1, 1),
                'cphase' : (gtype.U1, 1, 1),
                # U gate
                'u1' : (gtype.RZ, 0, 1), 'u2' : (gtype.U2, 0, 1), 'u3' : (gtype.U, 0, 1),
                # gtype of U1 gate is RZ instead of U1, for compatibility with OpenQASM.
                # controlled U gate
                'cu1' : (gtype.U1, 1, 1), 'cu2' : (gtype.U2, 1, 1), 'cu3' : (gtype.U, 1, 1),
                # swap and multi-controlled-bit gate
                'swap' : (gtype.SWAP, 0, 2),
                'ccx' : (gtype.X, 2, 1), 'ccz' : (gtype.Z, 2, 1)
            }

            if hasattr(qgate, '__version__') and qgate.__version__.startswith('0.4.0'):
                # register new gate types in Qgate 0.4
                QgateBackend.gatetypes['rxx'] = (gtype.RXX, 0, 2)
                QgateBackend.gatetypes['ryy'] = (gtype.RYY, 0, 2)
                QgateBackend.gatetypes['rzz'] = (gtype.RZZ, 0, 2)
                QgateBackend.gatetypes['cphase'] = (gtype.P, 1, 1)
                QgateBackend.gatetypes['cswap'] = (gtype.SWAP, 1, 2)
                # wrappers to absorb interface changes for Qgate 0.4
                def set_qregs_04(gate, qreg, controls):
                    gate.set_target(qreg)
                    if controls:
                        gate.set_controls(controls)

                def get_target_04(gate):
                    return gate.target

                def create_multi_qubit_gate_04(gate, typeinfo, targets):
                    g = model.Gate(typeinfo[0](*gate.params))
                    if ctrls:
                        q.set_controls(ctrls)
                    g.set_targets(targets)
                    return g

                def create_simulator_04(prefs):
                    return qgate.simulator.create(**prefs)

                def create_sampler_04(sim, qreg_ordering, sampling_method):
                    if sampling_method == 'blueqat' :
                        # create prob numpy array
                        probs = sim.qubits.create_prob_accessor(qreg_ordering)[:]
                        return BlueqatCompatibleSampler_04(probs, qreg_ordering)
                    else:
                        return SamplingWrapper_04(sim.qubits, qreg_ordering)

                QgateBackend.set_qregs = set_qregs_04
                QgateBackend.get_target = get_target_04
                QgateBackend.create_multi_qubit_gate = create_multi_qubit_gate_04
                QgateBackend.create_simulator = create_simulator_04
                QgateBackend.create_sampler = create_sampler_04
            else:
                # wrappers to absorb interface changes for Qgate 0.3 or older
                def set_qregs_03(gate, qreg, controls):
                    gate.set_qreg(qreg)
                    if controls:
                        gate.set_ctrllist(controls)

                def get_target_03(gate):
                    return gate.qreg

                def create_multi_qubit_gate_03(gate, typeinfo, ctrls, targets):
                    assert not ctrls, 'controlled multi qubit gate is not supported'
                    g = model.MultiQubitGate(typeinfo[0](*gate.params))
                    g.set_qreglist(targets)
                    return g

                def create_simulator_03(prefs):
                    circuit_prep = prefs.get('circuit_prep', qgate.prefs.dynamic)
                    simprefs = {'circuit_prep': circuit_prep}
                    # create simulator instance
                    runtime = prefs.get('runtime', 'cpu')
                    if runtime == 'cuda' :
                        sim = qgate.simulator.cuda(**simprefs)
                    elif runtime == 'cpu' :
                        sim = qgate.simulator.cpu(**simprefs)
                    elif runtime == 'py' :
                        sim = qgate.simulator.py(**simprefs)
                    else :
                        raise RuntimeError("Unknown runtime, {}.  Accetable values are 'cpu', 'cuda' and 'py'.".format(runtime))
                    return sim

                def create_sampler_03(sim, qreg_ordering, sampling_method):
                    if sampling_method == 'blueqat' :
                        # create prob numpy array
                        return sim.qubits.create_sampling_pool(qreg_ordering, BlueqatCompatibleSampler_03)
                    else:
                        return sim.qubits.create_sampling_pool(qreg_ordering)

                QgateBackend.set_qregs = set_qregs_03
                QgateBackend.get_target = get_target_03
                QgateBackend.create_multi_qubit_gate = create_multi_qubit_gate_03
                QgateBackend.create_simulator = create_simulator_03
                QgateBackend.create_sampler = create_sampler_03

    def run(self, gates, n_qubits, *args, **kwargs):
        self.n_qubits = n_qubits
        # returns
        r = kwargs.get('returns', '')
        sampling = kwargs.get('sampling', 'qgate')
        shots = kwargs.get('shots', 0)
        if shots != 0 and r == '' :
            r = 'shots'
        elif r == '' :
            r = 'statevector'

        if 'returns' in kwargs:
            del kwargs['returns']
        if 'sampling' in kwargs:
            del kwargs['sampling']
        if 'shots' in kwargs:
            del kwargs['shots']

        sim = QgateBackend.create_simulator(kwargs)

        # creating registers and references.
        self.qregs = [model.Qreg() for _ in range(n_qubits)]
        self.refs = [model.Reference() for _ in range(n_qubits)]
        sim.qubits.set_ordering(self.qregs)

        ops = self.convert_operators(gates)
        # qgate.dump(ops)
        if r == 'statevector' :
            return self.get_state_vector(sim, ops)
        elif r == 'shots' :
            return Counter(self.sample(sim, ops, shots, sampling))
        elif r == 'statevector_and_shots' :
            if 1 < shots :
                raise RuntimeError('shots must be 1 when returns="statevector_and_shots"')
            vec, shots = self.get_state_vector_and_sample(sim, ops)
            return vec, Counter(shots)
        else :
            raise RuntimeError('Unkown returns token, {}.'.format(r))

    def convert_operators(self, gates) :
        ops = list()
        for op in gates :
            typeinfo = QgateBackend.gatetypes.get(op.lowername, None)
            if typeinfo is not None :
                if typeinfo[2] == 1 :
                    ops += self.convert_one_qubit_gate(op, typeinfo)
                else :
                    ops += self.convert_multi_qubit_gate(op, typeinfo)
            elif op.lowername in ('measure', 'm') :
                # "measure": gate.Measurement, "m": gate.Measurement,
                ops += self.convert_measure(op)
            else :
                assert False
        return ops

    def convert_one_qubit_gate(self, gate, typeinfo) :
        glist = list()

        if typeinfo[1] == 0 :
            # no control bit
            assert isinstance(gate, bqgate.OneQubitGate)
            one_qubit_gates = self.create_gates(typeinfo[0], gate)
            # "tdg": gate.TDagGate, "sdg": gate.SDagGate,
            if gate.lowername[1:] == 'dg' :
                for qg in one_qubit_gates :
                    qg.set_adjoint(True)
            glist += one_qubit_gates
        else :
            if isinstance(gate, bqgate.TwoQubitGate) :
                # controlled gate
                for ctrlidx, qregidx in gate.control_target_iter(self.n_qubits) :
                    gt = typeinfo[0](*gate.params)
                    g = model.Gate(gt)
                    ctrl = self.qregs[ctrlidx]
                    qreg = self.qregs[qregidx]
                    QgateBackend.set_qregs(g, qreg, [ctrl])
                    glist.append(g)
            elif isinstance(gate, bqgate.ToffoliGate) :
                gt = typeinfo[0](*gate.params)
                g = model.Gate(gt)
                c0, c1, t = gate.targets
                ctrls = (self.qregs[c0], self.qregs[c1])
                qreg = self.qregs[t]
                QgateBackend.set_qregs(g, qreg, ctrls)
                glist.append(g)
            else :
                raise RuntimeError('Unknown gate, {}.'.format(repr(gate)))

        return glist

    def convert_multi_qubit_gate(self, gate, typeinfo) :
        glist = list()
        n_ctrls = typeinfo[1]
        ctrls = gate.targets[:n_ctrls]
        ctrl_qregs = (self.qregs[idx] for idx in ctrls)
        targets = gate.targets[n_ctrls]
        target_qregs = (self.qregs[idx] for idx in targets)
        g = QgateBackend.create_multi_qubit_gate(gate, typeinfo, ctrls, targets)
        glist.append(g)
        return glist

    def convert_measure(self, measure) :
        mlist = list()
        for qregidx in measure.target_iter(self.n_qubits) :
            m = model.Measure(self.refs[qregidx], self.qregs[qregidx])
            mlist.append(m)
        return mlist

    def create_gates(self, gate_type_factory, gate) :
        glist = list()
        for qregidx in gate.target_iter(self.n_qubits) :
            gt = gate_type_factory(*gate.params)
            g = model.Gate(gt)
            qreg = self.qregs[qregidx]
            QgateBackend.set_qregs(g, qreg, None)
            glist.append(g)
        return glist

    def create_global_phase_gates(self, phase, gate) :
        glist = list()
        for qregidx in gate.target_iter(self.n_qubits) :
            gt = gtype.ExpiI(phase)
            g = model.Gate(gt)
            qreg = self.qregs[qregidx]
            self.set_qregs(g, qreg, None)
            glist.append(g)
        return glist

    def get_state_vector(self, sim, ops) :
        sim.run(ops)
        return sim.qubits.states[:]

    def sample(self, sim, ops, shots, sampling_method) :
        sampling_qregs = list()
        ops_no_measure = list()

        for op in ops :
            if isinstance(op, model.Measure) :
                target = QgateBackend.get_target(op)
                sampling_qregs.append(target)
            else :
                ops_no_measure.append(op)

        sim.run(ops_no_measure)
        # print(sim.qubits.states[:])

        qregs_no_dup = set(sampling_qregs)
        sampling_qregs = list(qregs_no_dup)
        sampling_qregs.sort(key = self.qregs.index)

        sampler = QgateBackend.create_sampler(sim, sampling_qregs, sampling_method)
        obs = sampler.sample(shots)

        # FIXME: faster conversion for bit representation
        hist = obs.histgram()
        bit_format = '0{}b'.format(len(self.qregs))
        sampling_pos = [self.qregs.index(qreg) for qreg in sampling_qregs]
        max_bit_pos = len(self.qregs) - 1
        strkey_hist = dict()
        for v, c in hist.items() :
            permuted = 0
            for idx, pos in enumerate(sampling_pos) :
                if (v & (1 << idx)) != 0 :
                    permuted |= 1 << (max_bit_pos - pos)
            strbits = format(permuted, bit_format)
            strkey_hist[strbits] = c
        return strkey_hist

    def get_state_vector_and_sample(self, sim, ops) :
        sim.run(ops)
        # state vector
        state_vector = sim.qubits.states[:]
        # shot
        reversed_refs = list(reversed(self.refs))
        obs = sim.obs(reversed_refs)
        bit_format = '0{}b'.format(len(self.refs))
        strkey_hist = dict()
        strbits = format(obs.int, bit_format)
        strkey_hist[strbits] = 1 # observed 1 time.
        return state_vector, strkey_hist

# provided for compatiblity and tests
class BlueqatCompatibleSampler_03 :
    def __init__(self, prob, empty_lanes, qreg_ordering) :
        #prob *= (1. / np.sum(prob))
        self.prob = prob

        self.qreg_ordering = qreg_ordering
        self.empty_lanes = empty_lanes
        self.mask = 0
        for idx in empty_lanes :
            self.mask |= 1 << idx

    def sample(self, n_samples) :
        obs = np.empty([n_samples], np.int)
        import random
        for idx in range(n_samples) :
            redprob = self.prob
            v = 0
            p_all = 1
            for bitpos in range(len(self.qreg_ordering)) :
                rnum = random.random()
                p_zero = np.sum(redprob[::2])
                if rnum < (p_zero / p_all) :
                    redprob = redprob[::2]
                    p_all = p_zero
                else :
                    redprob = redprob[1::2]
                    p_all = p_all - p_zero
                    v |= (1 << bitpos)

            obs[idx] = v

        if self.mask != 0 :
            self.shift_for_empty_lanes(obs)
        obslist = qgate.simulator.observation.ObservationList(self.qreg_ordering, obs, self.mask)
        return obslist

    def shift_for_empty_lanes(self, obs) :
        # create control bit mask
        bits = [1 << lane for lane in self.empty_lanes]
        n_shifts = len(bits) + 1

        mask = bits[0] - 1
        masks = [ mask ]
        for idx in range(len(bits) - 1) :
            mask = (bits[idx + 1] - 1) & (~(bits[idx] * 2 - 1))
            masks.append(mask)
        mask = ~(bits[-1] * 2 - 1)
        masks.append(mask)

        for idx in range(len(obs)) :
            v = obs[idx]
            v_shifted = 0
            for shift in range(n_shifts) :
                v_shifted |= (v << shift) & masks[shift]
            obs[idx] = v_shifted


# provided for compatiblity and tests
class BlueqatCompatibleSampler_04 :
    def __init__(self, prob, qreg_ordering) :
        self.prob = prob
        self.qreg_ordering = qreg_ordering

    def sample(self, n_samples) :
        obs = np.empty([n_samples], np.int)
        import random
        for idx in range(n_samples) :
            redprob = self.prob
            v = 0
            p_all = 1
            for bitpos in range(len(self.qreg_ordering)) :
                rnum = random.random()
                p_zero = np.sum(redprob[::2])
                if rnum < (p_zero / p_all) :
                    redprob = redprob[::2]
                    p_all = p_zero
                else :
                    redprob = redprob[1::2]
                    p_all = p_all - p_zero
                    v |= (1 << bitpos)

            obs[idx] = v

        obslist = qgate.simulator.observation.ObservationList(self.qreg_ordering, obs, 0)
        return obslist


class SamplingWrapper_04:
    def __init__(self, qubits, qreg_ordering):
        self.qubits = qubits
        self.qreg_ordering = qreg_ordering

    def sample(self, n_samples, randnums=None):
        return self.qubits.sample(self.qreg_ordering, n_samples, randnums)



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cmath
import math
import random
import warnings
from collections import Counter

import numpy as np
from numba import jit, njit, prange
import numba

from ..gate import *
from .backendbase import Backend

DEFAULT_DTYPE = complex
FASTMATH = True

# Typedef
# Index of Qubit
_QBIdx = numba.uint32
_QBIdx_dtype = np.uint32
# Index of Quantum State
_QSIdx = numba.uint64
_QSIdx_dtype = np.uint64
# Mask of Quantum State
_QSMask = _QSIdx
_QSMask_dtype = _QSIdx_dtype

class _NumbaBackendContext:
    """This class is internally used in NumbaBackend"""

    def __init__(self, n_qubits, save_cache, dtype=DEFAULT_DTYPE):
        self.n_qubits = n_qubits
        self.qubits = np.zeros(2**n_qubits, dtype)
        self.save_cache = save_cache
        self.shots_result = Counter()
        self.cregs = None

    def prepare(self, cache):
        """Prepare to run next shot."""
        if cache is not None:
            np.copyto(self.qubits, cache)
        else:
            self.qubits.fill(0.0)
            self.qubits[0] = 1.0
        self.cregs = [0] * self.n_qubits

    def store_shot(self):
        """Store current cregs to shots_result"""
        def to_str(cregs):
            return ''.join(str(b) for b in cregs)
        key = to_str(self.cregs)
        self.shots_result[key] = self.shots_result.get(key, 0) + 1


@njit(_QSIdx(_QSMask, _QSIdx),
      locals={'lower': _QSMask, 'higher': _QSMask},
      nogil=True, cache=True)
def _shifted(lower_mask, idx):
    lower = idx & lower_mask
    higher = (idx & ~lower_mask) << 1
    return higher + lower


@njit(_QSMask[:](_QBIdx[:]), nogil=True, cache=True)
def _create_masks(indices):
    indices.sort()
    masks = np.empty(len(indices) + 1, dtype=_QSMask_dtype)
    for i, x in enumerate(indices):
        masks[i] = (1 << (x - i)) - 1
    masks[-1] = ~0
    for i in range(len(indices), 0, -1):
        masks[i] &= ~masks[i - 1]
    return masks


@njit(_QSIdx(_QSMask[:], _QSIdx), nogil=True, cache=True)
def _mult_shifted(masks, idx):
    shifted = 0
    for i, x in enumerate(masks):
        shifted |= (idx & x) << i
    return shifted


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _zgate(qubits, n_qubits, target):
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        qubits[_shifted(lower_mask, i) + (1 << target)] *= -1


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _xgate(qubits, n_qubits, target):
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        qubits[i0] = qubits[i0 + (1 << target)]
        qubits[i0 + (1 << target)] = t


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _ygate(qubits, n_qubits, target):
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0] * 1j
        qubits[i0] = qubits[i0 + (1 << target)] * -1j
        qubits[i0 + (1 << target)] = t


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _hgate(qubits, n_qubits, target):
    sqrt2_inv = 0.7071067811865475
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = (t + u) * sqrt2_inv
        qubits[i0 + (1 << target)] = (t - u) * sqrt2_inv


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _diaggate(qubits, n_qubits, target, factor):
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i1 = _shifted(lower_mask, i) + (1 << target)
        qubits[i1] *= factor


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _rzgate(qubits, n_qubits, target, ang):
    ang *= 0.5
    eit = cmath.exp(1.j * ang)
    eitstar = eit.conjugate()
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] *= eitstar
        qubits[i0 + (1 << target)] *= eit


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _rygate(qubits, n_qubits, target, ang):
    ang *= 0.5
    cos = math.cos(ang)
    sin = math.sin(ang)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = cos * t - sin * u
        qubits[i0 + (1 << target)] = sin * t + cos * u


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _rxgate(qubits, n_qubits, target, ang):
    ang *= 0.5
    cos = math.cos(ang)
    nisin = math.sin(ang) * -1.j
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = cos * t + nisin * u
        qubits[i0 + (1 << target)] = nisin * t + cos * u


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _u3gate(qubits, n_qubits, target, theta, phi, lambd):
    theta *= 0.5
    cos = math.cos(theta)
    sin = math.sin(theta)
    expadd = cmath.exp((phi + lambd) * 0.5j)
    expsub = cmath.exp((phi - lambd) * 0.5j)
    a = expadd.conjugate() * cos
    b = -expsub.conjugate() * sin
    c = expsub * sin
    d = expadd * cos
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = a * t + b * u
        qubits[i0 + (1 << target)] = c * t + d * u

@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _mat1gate(qubits, n_qubits, target, mat):
    lower_mask = (1 << _QSMask(target)) - 1
    a = mat[0, 0]
    b = mat[0, 1]
    c = mat[1, 0]
    d = mat[1, 1]
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        t = qubits[i0]
        u = qubits[i0 + (1 << target)]
        qubits[i0] = a * t + b * u
        qubits[i0 + (1 << target)] = c * t + d * u


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _czgate(qubits, n_qubits, controls_target):
    target = controls_target[-1]
    all1 = _QSMask(0)
    for b in controls_target:
        all1 |= _QSMask(1) << b
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i11 = _mult_shifted(masks, i) | all1
        qubits[i11] *= -1


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _cxgate(qubits, n_qubits, controls_target):
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        t = qubits[i10]
        qubits[i10] = qubits[i11]
        qubits[i11] = t


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _crxgate(qubits, n_qubits, controls_target, ang):
    ang *= 0.5
    cos = math.cos(ang)
    nisin = math.sin(ang) * -1.j
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        t = qubits[i10]
        u = qubits[i11]
        qubits[i10] = cos * t + nisin * u
        qubits[i11] = nisin * t + cos * u


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _crygate(qubits, n_qubits, controls_target, ang):
    ang *= 0.5
    cos = math.cos(ang)
    sin = math.sin(ang)
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        t = qubits[i10]
        u = qubits[i11]
        qubits[i10] = cos * t - sin * u
        qubits[i11] = sin * t + cos * u


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _crzgate(qubits, n_qubits, controls_target, ang):
    ang *= 0.5
    eit = cmath.exp(1.j * ang)
    eitstar = eit.conjugate()
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i10 = _mult_shifted(masks, i) | c_mask
        i11 = i10 | t_mask
        qubits[i10] *= eitstar
        qubits[i11] *= eit


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _cphasegate(qubits, n_qubits, controls_target, ang):
    eit = cmath.exp(1.j * ang)
    c_mask = _QSMask(0)
    for c in controls_target[:-1]:
        c_mask |= _QSMask(1) << c
    t_mask = 1 << controls_target[-1]
    n_loop = 1 << (_QSMask(n_qubits) - _QSMask(len(controls_target)))
    masks = _create_masks(controls_target)
    for i in prange(n_loop):
        i11 = _mult_shifted(masks, i) | c_mask | t_mask
        qubits[i11] *= eit


@njit(locals={'lower_mask': _QSMask},
      nogil=True, parallel=True, fastmath=FASTMATH)
def _p0calc(qubits, target, n_qubits):
    p0 = 0.0
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        val = qubits[_shifted(lower_mask, i)]
        p0 += val.real * val.real + val.imag * val.imag
    return p0


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _reduce0(qubits, target, n_qubits, p0):
    sqrtp_inv = 1.0 / math.sqrt(p0)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        qubits[i0] *= sqrtp_inv
        qubits[i0 + (1 << target)] = 0.0


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _reduce1(qubits, target, n_qubits, p0):
    sqrtp_inv = 1.0 / math.sqrt(1.0 - p0)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        qubits[i0 + (1 << target)] *= sqrtp_inv
        qubits[i0] = 0.0


@njit(nogil=True, parallel=True, fastmath=FASTMATH)
def _reset1(qubits, target, n_qubits, p0):
    sqrtp_inv = 1.0 / math.sqrt(1.0 - p0)
    lower_mask = (1 << _QSMask(target)) - 1
    for i in prange(1 << (_QSMask(n_qubits) - 1)):
        i0 = _shifted(lower_mask, i)
        qubits[i0] = qubits[i0 + (1 << target)] * sqrtp_inv
        qubits[i0 + (1 << target)] = 0.0


class NumbaBackend(Backend):
    """Simulator backend which uses numba."""
    __return_type = {
        "statevector": lambda ctx: ctx.qubits,
        "shots": lambda ctx: ctx.shots_result,
        "statevector_and_shots": lambda ctx: (ctx.qubits, ctx.shots_result),
        "_inner_ctx": lambda ctx: ctx,
    }
    DEFAULT_SHOTS = 1024

    def __init__(self):
        self.cache = None
        self.cache_idx = -1

    def __clear_cache(self):
        self.cache = None
        self.cache_idx = -1

    def __clear_cache_if_invalid(self, n_qubits, dtype):
        if self.cache is None:
            self.__clear_cache()
            return
        if len(self.cache) != 2**n_qubits:
            self.__clear_cache()
            return
        if self.cache.dtype != dtype:
            self.__clear_cache()
            return

    def run(self, gates, n_qubits, *args, **kwargs):
        def __parse_run_args(shots=None, returns=None, enable_cache=True, ignore_global=False,
                             dtype=DEFAULT_DTYPE, **_kwargs):
            if returns is None:
                if shots is None:
                    returns = "statevector"
                else:
                    returns = "shots"
            if returns not in self.__return_type.keys():
                raise ValueError(f"Unknown returns type '{returns}'")
            if shots is None:
                if returns in ("statevector", "_inner_ctx"):
                    shots = 1
                else:
                    shots = self.DEFAULT_SHOTS
            if returns == "statevector" and shots > 1:
                warnings.warn("When `returns` = 'statevector', `shots` = 1 is enough.")
            return shots, returns, dtype, enable_cache, ignore_global

        shots, returns, dtype, enable_cache, ignore_global = __parse_run_args(*args, **kwargs)

        if enable_cache:
            self.__clear_cache_if_invalid(n_qubits, dtype)
        else:
            self.__clear_cache()
        ctx = _NumbaBackendContext(n_qubits, enable_cache, dtype)

        def run_single_gate(gate):
            nonlocal ctx
            action = self._get_action(gate)
            if action is not None:
                ctx = action(gate, ctx)
            else:
                for g in gate.fallback(n_qubits):
                    run_single_gate(g)

        for _ in range(shots):
            ctx.prepare(self.cache)
            cache_idx = self.cache_idx
            for gate in gates[cache_idx + 1:]:
                run_single_gate(gate)
                if ctx.save_cache:
                    cache_idx += 1
            if self.cache_idx != cache_idx:
                self.cache_idx = cache_idx
                if ctx.save_cache:
                    self.cache = ctx.qubits.copy()
            if ctx.cregs:
                ctx.store_shot()

        if ignore_global:
            _ignore_global(ctx.qubits)
        return self.__return_type[returns](ctx)

    def make_cache(self, gates, n_qubits):
        self.run(gates, n_qubits)

    def gate_x(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _xgate(qubits, n_qubits, target)
        return ctx

    def gate_y(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _ygate(qubits, n_qubits, target)
        return ctx

    def gate_z(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _zgate(qubits, n_qubits, target)
        return ctx

    def gate_h(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            _hgate(qubits, n_qubits, target)
        return ctx

    def gate_t(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = cmath.exp(0.25j * math.pi)
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    def gate_tdg(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = cmath.exp(-0.25j * math.pi)
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    def gate_s(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = 1.j
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    def gate_sdg(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = -1.j
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    def gate_cz(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for control, target in gate.control_target_iter(n_qubits):
            _czgate(qubits, n_qubits, np.array([control, target], dtype=_QBIdx_dtype))
        return ctx

    def gate_cx(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for control, target in gate.control_target_iter(n_qubits):
            _cxgate(qubits, n_qubits, np.array([control, target], dtype=_QBIdx_dtype))
        return ctx

    def gate_rx(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for target in gate.target_iter(n_qubits):
            _rxgate(qubits, n_qubits, target, theta)
        return ctx

    def gate_ry(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for target in gate.target_iter(n_qubits):
            _rygate(qubits, n_qubits, target, theta)
        return ctx

    def gate_rz(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for target in gate.target_iter(n_qubits):
            _rzgate(qubits, n_qubits, target, theta)
        return ctx

    def gate_phase(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        factor = cmath.exp(1.j * gate.theta)
        for target in gate.target_iter(n_qubits):
            _diaggate(qubits, n_qubits, target, factor)
        return ctx

    def gate_crx(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _crxgate(qubits, n_qubits, np.array([control, target], dtype=_QBIdx_dtype), theta)
        return ctx

    def gate_cry(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _crygate(qubits, n_qubits, np.array([control, target], dtype=_QBIdx_dtype), theta)
        return ctx

    def gate_crz(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _crzgate(qubits, n_qubits, np.array([control, target], dtype=_QBIdx_dtype), theta)
        return ctx

    def gate_cphase(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        for control, target in gate.control_target_iter(n_qubits):
            _cphasegate(qubits, n_qubits, np.array([control, target], dtype=_QBIdx_dtype), theta)
        return ctx

    def gate_ccz(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        _czgate(qubits, n_qubits, np.array(gate.targets, dtype=_QBIdx_dtype))
        return ctx

    def gate_ccx(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        _cxgate(qubits, n_qubits, np.array(gate.targets, dtype=_QBIdx_dtype))
        return ctx

    def gate_u1(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        angle = gate.lambd
        for target in gate.target_iter(n_qubits):
            _rzgate(qubits, n_qubits, target, angle)
        return ctx

    def gate_u3(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        theta = gate.theta
        phi = gate.phi
        lambd = gate.lambd
        for target in gate.target_iter(n_qubits):
            _u3gate(qubits, n_qubits, target, theta, phi, lambd)
        return ctx

    def gate_mat1(self, gate, ctx):
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        mat = gate.matrix()
        for target in gate.target_iter(n_qubits):
            _mat1gate(qubits, n_qubits, target, mat)
        return ctx

    def gate_measure(self, gate, ctx):
        if ctx.save_cache:
            self.cache = ctx.qubits.copy()
        ctx.save_cache = False
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            rand = random.random()
            p0 = _p0calc(qubits, target, n_qubits)
            if rand < p0:
                _reduce0(qubits, target, n_qubits, p0)
                ctx.cregs[target] = 0
            else:
                _reduce1(qubits, target, n_qubits, p0)
                ctx.cregs[target] = 1
        return ctx

    def gate_reset(self, gate, ctx):
        if ctx.save_cache:
            self.cache = ctx.qubits.copy()
        ctx.save_cache = False
        qubits = ctx.qubits
        n_qubits = ctx.n_qubits
        for target in gate.target_iter(n_qubits):
            rand = random.random()
            p0 = _p0calc(qubits, target, n_qubits)
            if rand < p0:
                _reduce0(qubits, target, n_qubits, p0)
            else:
                _reset1(qubits, target, n_qubits, p0)
        return ctx


@njit(nogil=True, cache=True)
def _ignore_global(qubits):
    for q in qubits:
        if abs(q) > 0.0000001:
            ang = abs(q) / q
            qubits *= ang
            break
    return qubits



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import cmath
from typing import List

import numpy as np

from ..utils import check_unitarity
from ..gate import *


def u3_decomposer(gate: OneQubitGate) -> List[U3Gate]:
    """Decompose one qubit gate to U3 gate.
    This decomposer ignores global phase.

    Args:
        gate (OneQubitGate): The gate which have 2x2 unitary matrix

    Returns:
        List of U3 gate. length is always 1.
    """
    mat = gate.matrix()
    assert mat.shape == (2, 2)
    assert check_unitarity(mat)
    # Remove global phase. Make SU(2).
    mat /= np.sqrt(np.linalg.det(mat))
    sq_cos_halftheta = (mat[0, 0] * mat[1, 1]).real
    cos_halftheta = np.sqrt(sq_cos_halftheta)
    sin_halftheta = np.sqrt(1. - sq_cos_halftheta)
    theta = np.arccos(cos_halftheta) * 2
    if np.allclose(0.0, cos_halftheta):
        lam = 0.0
        phi = cmath.phase(mat[1, 0]) * 2
    elif np.allclose(0.0, sin_halftheta):
        lam = 0.0
        phi = cmath.phase(mat[1, 1]) * 2
    else:
        angle_phi_plus_lambda_half = cmath.phase(mat[1, 1] / sin_halftheta)
        angle_phi_minus_lambda_half = cmath.phase(mat[1, 0] / cos_halftheta)
        phi = angle_phi_plus_lambda_half + angle_phi_minus_lambda_half
        lam = angle_phi_plus_lambda_half - angle_phi_minus_lambda_half
    return [U3Gate(gate.targets, theta, phi, lam)]



# Copyright 2019 The Blueqat Developers
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import numpy as np
from blueqat import Circuit

def circuit_to_unitary(circ: Circuit, *runargs, **runkwargs):
    runkwargs.setdefault('returns', 'statevector')
    runkwargs.setdefault('ignore_global', False)
    n_qubits = circ.n_qubits
    vecs = []
    if n_qubits == 0:
        return np.array([[1]])
    for i in range(1 << n_qubits):
        bitmask = tuple(k for k in range(n_qubits) if (1 << k) & i)
        c = Circuit()
        if bitmask:
            c.x[bitmask]
        c += circ
        vecs.append(c.run(*runargs, **runkwargs))
    return np.array(vecs).T



from typing import Tuple, NamedTuple
from functools import partial
from .. import gate
from .. import Circuit

GATE_SET = {
    "i": gate.IGate,
    "x": gate.XGate,
    "y": gate.YGate,
    "z": gate.ZGate,
    "h": gate.HGate,
    "t": gate.TGate,
    "tdg": gate.TDagGate,
    "s": gate.SGate,
    "sdg": gate.SDagGate,
    "cz": gate.CZGate,
    "cx": gate.CXGate,
    "cnot": gate.CXGate,
    "cy": gate.CYGate,
    "ch": gate.CHGate,
    "rx": gate.RXGate,
    "ry": gate.RYGate,
    "rz": gate.RZGate,
    "r": gate.PhaseGate,
    "phase": gate.PhaseGate,
    "crx": gate.CRXGate,
    "cry": gate.CRYGate,
    "crz": gate.CRZGate,
    "cr": gate.CPhaseGate,
    "cphase": gate.CPhaseGate,
    "u1": gate.U1Gate,
    "u2": gate.U2Gate,
    "u3": gate.U3Gate,
    "cu1": gate.CU1Gate,
    "cu2": gate.CU2Gate,
    "cu3": gate.CU3Gate,
    "swap": gate.SwapGate,
    "ccx": gate.ToffoliGate,
    "toffoli": gate.ToffoliGate,
    "measure": gate.Measurement,
    "m": gate.Measurement,
}

GLOBAL_MACROS = {}

class Ops(NamedTuple):
    """Immutable type operations"""
    n_qubits: int = 0
    ops: Tuple[gate.Gate] = ()

    def __repr__(self):
        return f'Ops({self.n_qubits}).' + '.'.join(str(op) for op in self.ops)

    def __getattr__(self, name):
        if name in GATE_SET:
            return _GateWrapper(self, name, GATE_SET[name])
        if name in GLOBAL_MACROS:
            return partial(GLOBAL_MACROS[name], self)
        raise AttributeError(f"'circuit' object has no attribute or gate '{name}'")

    def __add__(self, other):
        if not isinstance(other, Ops):
            return NotImplemented
        return Ops(max(self.n_qubits, other.n_qubits), self.ops + other.ops)

    def dagger(self):
        """Get Hamiltonian conjugate of `self`."""
        return Ops(self.n_qubits, tuple(g.dagger() for g in reversed(self.ops)))

    def to_circuit(self):
        return Circuit(self.n_qubits, list(self.ops))


class _GateWrapper:
    def __init__(self, ops, name, gate):
        self.ops = ops
        self.target = None
        self.name = name
        self.gate = gate
        self.args = ()
        self.kwargs = {}

    def __call__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        return self

    def __getitem__(self, args):
        self.target = args
        n_qubits = max(self.ops.n_qubits, gate.get_maximum_index(args) + 1)
        ops = self.ops.ops + (self.gate(self.target, *self.args, **self.kwargs),)
        return Ops(n_qubits, ops)

    def __str__(self):
        if self.args:
            args_str = str(self.args)
            if self.kwargs:
                args_str = args_str[:-1] + ", kwargs=" + str(self.kwargs) + ")"
        elif self.kwargs:
            args_str = "(kwargs=" + str(self.kwargs) + ")"
        else:
            args_str = ""
        return self.name + args_str + " " + str(self.target)



"""
`gate` module implements quantum gate operations.
This module is internally used.
"""

import math
from typing import List

import numpy as np


class Operation:
    """Abstract quantum circuit operation class."""

    """Lower name of the operation."""
    lowername: str = None

    @property
    def uppername(self) -> str:
        """Upper name of the gate."""
        return self.lowername.upper()

    def __init__(self, targets, params=(), **kwargs):
        if self.lowername is None:
            raise ValueError(f"{self.__class__.__name__}.lowername is not defined.")
        self.params = params
        self.kwargs = kwargs
        self.targets = targets

    def target_iter(self, n_qubits):
        """The generator which yields the target qubits."""
        return slicing(self.targets, n_qubits)

    def fallback(self, n_qubits: int) -> List['Operation']:
        """Returns alternative operations to make equivalent circuit."""
        raise NotImplementedError(f"The fallback of {self.__class__.__name__} operation is not defined.")

    def dagger(self) -> 'Operation':
        """Returns the Hermitian conjugate of `self`."""
        raise NotImplementedError(f"Hermitian conjugate of {self.__class__.__name__} operation is not provided.")

    def _str_args(self) -> str:
        """Returns printable string of args."""
        if not self.params:
            return ''
        return '(' + ', '.join(str(param) for param in self.params) + ')'

    def _str_targets(self) -> str:
        """Returns printable string of targets."""
        def _slice_to_str(obj):
            if isinstance(obj, slice):
                start = '' if obj.start is None else str(obj.start.__index__())
                stop = '' if obj.stop is None else str(obj.stop.__index__())
                if obj.step is None:
                    return f'{start}:{stop}'
                else:
                    step = str(obj.step.__index__())
                    return f'{start}:{stop}:{step}'
            else:
                return str(obj.__index__())

        if isinstance(self.targets, tuple):
            return f"[{', '.join(_slice_to_str(target) for target in self.targets)}]"
        else:
            return f"[{_slice_to_str(self.targets)}]"

    def __str__(self) -> str:
        str_args = self._str_args()
        str_targets = self._str_targets()
        return f'{self.lowername}{str_args}{str_targets}'


class Gate(Operation):
    """Abstract quantum gate class."""
    @property
    def n_qargs(self) -> int:
        """Number of qubit arguments of this gate."""
        raise NotImplementedError()

    def fallback(self, n_qubits: int) -> List['Gate']:
        """Returns alternative gates to make equivalent circuit."""
        raise NotImplementedError(f"The fallback of {self.__class__.__name__} gate is not defined.")

    def dagger(self) -> 'Gate':
        """Returns the Hermitian conjugate of `self`."""
        raise NotImplementedError(f"Hermitian conjugate of this gate is not provided.")

    def matrix(self) -> np.array:
        """Returns the matrix of implementations.

        (Non-abstract) subclasses of Gate must implement this method.
        WARNING: qubit order specifications of multi qubit gate is still not defined.
        """
        raise NotImplementedError()


class OneQubitGate(Gate):
    """Abstract quantum gate class for 1 qubit gate."""
    @property
    def n_qargs(self):
        return 1

    def _make_fallback_for_target_iter(self, n_qubits, fallback):
        gates = []
        for t in self.target_iter(n_qubits):
            gates += fallback(t)
        return gates


class TwoQubitGate(Gate):
    """Abstract quantum gate class for 2 qubits gate."""
    @property
    def n_qargs(self):
        return 2

    def control_target_iter(self, n_qubits):
        """The generator which yields the tuples of (control, target) qubits."""
        return qubit_pairs(self.targets, n_qubits)

    def _make_fallback_for_control_target_iter(self, n_qubits, fallback):
        gates = []
        for c, t in self.control_target_iter(n_qubits):
            gates += fallback(c, t)
        return gates


class IGate(OneQubitGate):
    """Identity gate"""
    lowername = "i"

    def fallback(self, n_qubits):
        return []

    def dagger(self):
        return self

    def matrix(self):
        return np.eye(2)


class XGate(OneQubitGate):
    """Pauli's X gate"""
    lowername = "x"

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[0, 1], [1, 0]], dtype=complex)


class YGate(OneQubitGate):
    """Pauli's Y gate"""
    lowername = "y"

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[0, -1j], [1j, 0]])


class ZGate(OneQubitGate):
    """Pauli's Z gate"""
    lowername = "z"

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[1, 0], [0, -1]], dtype=complex)


class HGate(OneQubitGate):
    """Hadamard gate"""
    lowername = "h"

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[1, 1], [1, -1]], dtype=complex) / np.sqrt(2)


class RXGate(OneQubitGate):
    """Rotate-X gate"""
    lowername = "rx"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return RXGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        t = self.theta * 0.5
        a = np.cos(t)
        b = -1j * np.sin(t)
        return np.array([[a, b], [b, a]], dtype=complex)


class RYGate(OneQubitGate):
    """Rotate-Y gate"""
    lowername = "ry"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return RYGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        t = self.theta * 0.5
        a = np.cos(t)
        b = np.sin(t)
        return np.array([[a, -b], [b, a]], dtype=complex)


class RZGate(OneQubitGate):
    """Rotate-Z gate"""
    lowername = "rz"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return RZGate(self.targets, -self.theta, **self.kwargs)

    def matrix(self):
        a = np.exp(0.5j * self.theta)
        return np.array([[a.conjugate(), 0], [0, a]], dtype=complex)


class PhaseGate(OneQubitGate):
    """Rotate-Z gate but global phase is different.

    Global phase doesn't makes any difference of measured result.
    You may use RZ gate or U1 gate instead, but distinguishing these gates
    may better for debugging or future improvement.

    furthermore, phase gate may efficient for simulating.
    (It depends on backend implementation.
    But matrix of phase gate is simpler than RZ gate or U1 gate.)"""
    lowername = "phase"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return PhaseGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        # If phase gate is not implemented in the backend, global phase is ignored.
        return self._make_fallback_for_target_iter(n_qubits, lambda t: [RZGate(t, self.theta)])

    def matrix(self):
        return np.array([[1, 0], [0, np.exp(self.theta)]], dtype=complex)


class TGate(OneQubitGate):
    """T ($\\pi/8$) gate"""
    lowername = "t"

    def dagger(self):
        return TDagGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(n_qubits, lambda t: [PhaseGate(t, math.pi / 4)])

    def matrix(self):
        return np.array([[1, 0], [0, np.exp(np.pi * 0.25)]])


class TDagGate(OneQubitGate):
    """Dagger of T ($\\pi/8$) gate"""
    lowername = "tdg"

    def dagger(self):
        return TGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(n_qubits, lambda t: [PhaseGate(t, -math.pi / 4)])

    def matrix(self):
        return np.array([[1, 0], [0, np.exp(np.pi * -0.25)]])


class SGate(OneQubitGate):
    """S gate"""
    lowername = "s"

    def dagger(self):
        return SDagGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(n_qubits, lambda t: [PhaseGate(t, math.pi / 2)])

    def matrix(self):
        return np.array([[1, 0], [0, 1j]])


class SDagGate(OneQubitGate):
    """Dagger of S gate"""
    lowername = "sdg"

    def dagger(self):
        return SGate(self.targets, self.params, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(n_qubits, lambda t: [PhaseGate(t, -math.pi / 2)])

    def matrix(self):
        return np.array([[1, 0], [0, -1j]])


class U1Gate(OneQubitGate):
    """U1 gate

    U1 gate is as same as RZ gate and CU1 gate is as same as CPhase gate.
    It is because for compatibility with IBM's implementations.

    You should probably use RZ/CRZ gates or Phase/CPhase gates instead of U1/CU1 gates.
    """
    lowername = "u1"

    def __init__(self, targets, lambd, **kwargs):
        super().__init__(targets, (lambd,), **kwargs)
        self.lambd = lambd

    def dagger(self):
        return U1Gate(self.targets, -self.lambd, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(
            n_qubits, lambda t: [U3Gate(t, 0.0, 0.0, self.lambd)])

    def matrix(self):
        a = np.exp(0.5j * self.theta)
        return np.array([[a.conjugate(), 0], [0, a]], dtype=complex)


class U2Gate(OneQubitGate):
    """U2 gate"""
    lowername = "u2"

    def __init__(self, targets, phi, lambd, **kwargs):
        super().__init__(targets, (phi, lambd), **kwargs)
        self.phi = phi
        self.lambd = lambd

    def dagger(self):
        return U3Gate(self.targets, -math.pi / 2, -self.lambd, -self.phi, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_target_iter(
            n_qubits, lambda t: [U3Gate(t, math.pi / 2, self.phi, self.lambd)])

    def matrix(self):
        p, l = self.params
        c = 1 / np.sqrt(2)
        a = np.exp(0.5j * (p + l)) * c
        b = np.exp(0.5j * (p - l)) * c
        return np.array([[a.conjugate(), -b.conjugate()], [a, b]], dtype=complex)


class U3Gate(OneQubitGate):
    """U3 gate"""
    lowername = "u3"

    def __init__(self, targets, theta, phi, lambd, **kwargs):
        super().__init__(targets, (theta, phi, lambd), **kwargs)
        self.theta = theta
        self.phi = phi
        self.lambd = lambd

    def dagger(self):
        return U3Gate(self.targets, -self.theta, -self.lambd, -self.phi, **self.kwargs)

    def matrix(self):
        t, p, l = self.params
        a = np.exp(0.5j * (p + l)) * np.cos(t * 0.5)
        b = np.exp(0.5j * (p - l)) * np.sin(t * 0.5)
        return np.array([[a.conjugate(), -b.conjugate()], [b, a]], dtype=complex)


class Mat1Gate(OneQubitGate):
    """Arbitrary 2x2 matrix gate

    `mat` is expected a 2x2 unitary matrix, but not checked.
    (If unexpected matrix is given, backend may raises error or returns weird result)
    """
    lowername = "mat1"

    def __init__(self, targets, mat: np.array, **kwargs):
        super().__init__(targets, (mat,), **kwargs)
        self.mat = mat

    def dagger(self):
        return Mat1Gate(self.targets, self.mat.T.conjugate(), **self.kwargs)

    def matrix(self):
        return self.mat


class CXGate(TwoQubitGate):
    """Controlled-X (CNOT) gate"""
    lowername = "cx"

    def dagger(self):
        return self

    # TODO: Specification required: target \otimes control or control \otimes target
    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 0, 0, 1], [0, 0, 1, 0], [0, 1, 0, 0]], dtype=complex)


class CZGate(TwoQubitGate):
    """Controlled-Z gate"""
    lowername = "cz"

    def dagger(self):
        return self

    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, -1]], dtype=complex)


class CYGate(TwoQubitGate):
    """Controlled-Y gate"""
    lowername = "cy"

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [SDagGate(t),
                          CXGate((c, t)),
                          SGate(t)])

    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 0, -1j, 0], [0, 0, 1, 0], [0, 1j, 0, 0]], dtype=complex)


class CHGate(TwoQubitGate):
    """Controlled-H gate"""
    lowername = "ch"

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        # Ignores global phase
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [RYGate(t, math.pi / 4),
                          CXGate((c, t)),
                          RYGate(t, -math.pi / 4)])

    def matrix(self):
        a = 1.0 / np.sqrt(2)
        return np.array([[1, 0, 0, 0], [0, a, 0, a], [0, 0, 1, 0], [0, a, 0, -a]], dtype=complex)


class CRXGate(TwoQubitGate):
    """Rotate-X gate"""
    lowername = "crx"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return CRXGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [RXGate(t, self.theta / 2),
                          CZGate((c, t)),
                          RXGate(t, -self.theta / 2),
                          CZGate((c, t))])

    def matrix(self):
        t = self.theta * 0.5
        a = np.cos(t)
        b = -1j * np.sin(t)
        return np.array([[1, 0, 0, 0], [0, a, 0, b], [0, 0, 1, 0], [0, b, 0, a]], dtype=complex)


class CRYGate(TwoQubitGate):
    """Rotate-Y gate"""
    lowername = "cry"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return CRYGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [RYGate(t, self.theta / 2),
                          CXGate((c, t)),
                          RYGate(t, -self.theta / 2),
                          CXGate((c, t))])

    def matrix(self):
        t = self.theta * 0.5
        a = np.cos(t)
        b = np.sin(t)
        return np.array([[1, 0, 0, 0], [0, a, 0, -b], [0, 0, 1, 0], [0, b, 0, a]], dtype=complex)


class CRZGate(TwoQubitGate):
    """Rotate-Z gate"""
    lowername = "crz"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return CRZGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [RZGate(t, self.theta / 2),
                          CXGate((c, t)),
                          RZGate(t, -self.theta / 2),
                          CXGate((c, t))])

    def matrix(self):
        a = np.exp(0.5j * self.theta)
        return np.array([[1, 0, 0, 0], [0, a.conjugate(), 0, 0], [0, 0, 1, 0], [0, 0, 0, a]], dtype=complex)


class CPhaseGate(TwoQubitGate):
    """Rotate-Z gate but phase is different."""
    lowername = "cphase"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return CPhaseGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [CRZGate((c, t), self.theta),
                          PhaseGate(c, self.theta / 2)])

    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0, np.exp(self.theta)]], dtype=complex)


class RXXGate(TwoQubitGate):
    """Rotate-XX gate"""
    lowername = "rxx"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return RXXGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        # TODO: test
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [HGate(c),
                          HGate(t),
                          RZZGate((c, t), self.theta),
                          HGate(c),
                          HGate(t)])

    def matrix(self):
        a = np.cos(self.theta * 0.5)
        b = -1j * np.sin(self.theta * 0.5)
        return np.array([[a, 0, 0, b], [0, a, b, 0], [0, b, a, 0], [a, 0, 0, b]], dtype=complex)


class RYYGate(TwoQubitGate):
    """Rotate-YY gate"""
    lowername = "ryy"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return RYYGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        # TODO: test
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [RXGate(c, -math.pi * 0.5),
                          RXGate(t, -math.pi * 0.5),
                          RZZGate((c, t), self.theta),
                          RXGate(c, math.pi * 0.5),
                          RXGate(t, math.pi * 0.5)])

    def matrix(self):
        a = np.cos(self.theta * 0.5)
        b = 1j * np.sin(self.theta * 0.5)
        return np.array([[a, 0, 0, b], [0, a, -b, 0], [0, -b, a, 0], [a, 0, 0, b]], dtype=complex)


class RZZGate(TwoQubitGate):
    """Rotate-ZZ gate"""
    lowername = "rzz"

    def __init__(self, targets, theta, **kwargs):
        super().__init__(targets, (theta,), **kwargs)
        self.theta = theta

    def dagger(self):
        return RZZGate(self.targets, -self.theta, **self.kwargs)

    def fallback(self, n_qubits):
        # TODO: test
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [CXGate((c, t)),
                          RZGate(t, self.theta),
                          CXGate((c, t))])

    def matrix(self):
        a = np.exp(0.5j * self.theta)
        return np.array([[a.conjugate(), 0, 0, 0], [0, a, 0, 0], [0, 0, a, 0], [0, 0, 0, a.conjugate()]], dtype=complex)


class SwapGate(TwoQubitGate):
    """Swap gate"""
    lowername = "swap"

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits,
            lambda c, t: [CXGate((c, t)), CXGate((t, c)), CXGate((c, t))])

    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


class ZZGate(TwoQubitGate):
    """ZZ gate

    This gate is a basis two-qubit gate for some kinds of trapped-ion based machines.
    It is equivalent with RZZ(pi/2) except global phase.
    """
    lowername = "zz"

    def __init__(self, targets, **kwargs):
        super().__init__(targets, (), **kwargs)

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        # Ignoring global phase.
        return self._make_fallback_for_control_target_iter(
                n_qubits, lambda c, t: [
                    RYGate(t, math.pi * 0.5),
                    CXGate((c, t)),
                    RZGate(c, math.pi * 0.5),
                    U3Gate(t, -math.pi * 0.5, math.pi * 0.5, 0),
                ])

        def matrix(self):
            return np.diag([1, 1j, 1j, 1])


class CU1Gate(TwoQubitGate):
    """Controlled U1 gate

    U1 gate is as same as RZ gate and CU1 gate is as same as CPhase gate.
    It is because for compatibility with IBM's implementations.

    You should probably use RZ/CRZ gates or Phase/CPhase gates instead of U1/CU1 gates.
    """
    lowername = "cu1"

    def __init__(self, targets, lambd, **kwargs):
        super().__init__(targets, (lambd,), **kwargs)
        self.lambd = lambd

    def dagger(self):
        return CU1Gate(self.targets, -self.lambd, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t: [
                U1Gate(c, self.lambd / 2),
                CXGate((c, t)),
                U1Gate(t, -self.lambd / 2),
                CXGate((c, t)),
                U1Gate(t, self.lambd / 2),
            ])

    def matrix(self):
        return np.array([[1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 0, np.exp(self.theta)]], dtype=complex)


class CU2Gate(TwoQubitGate):
    """Controlled U2 gate"""
    lowername = "cu2"

    def __init__(self, targets, phi, lambd, **kwargs):
        super().__init__(targets, (phi, lambd), **kwargs)
        self.phi = phi
        self.lambd = lambd

    def dagger(self):
        return CU3Gate(self.targets, -math.pi / 2, -self.lambd, -self.phi, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t: [CU3Gate((c, t), math.pi / 2, self.phi, self.lambd)])

    def matrix(self):
        p, l = self.params
        c = 1 / np.sqrt(2)
        a = np.exp(0.5j * (p + l)) * c
        b = np.exp(0.5j * (p - l)) * c
        return np.array([[1, 0, 0, 0], [0, a.conjugate(), 0, -b.conjugate()], [0, 0, 1, 0], [0, a, 0, b]], dtype=complex)


class CU3Gate(TwoQubitGate):
    """Controlled U3 gate"""
    lowername = "cu3"

    def __init__(self, targets, theta, phi, lambd, **kwargs):
        super().__init__(targets, (theta, phi, lambd), **kwargs)
        self.theta = theta
        self.phi = phi
        self.lambd = lambd

    def dagger(self):
        return CU3Gate(self.targets, -self.theta, -self.lambd, -self.phi, **self.kwargs)

    def fallback(self, n_qubits):
        return self._make_fallback_for_control_target_iter(
            n_qubits, lambda c, t: [
                U1Gate(t, (self.lambd - self.phi) / 2),
                CXGate((c, t)),
                U3Gate(t, -self.theta / 2, 0, -(self.phi + self.lambd) / 2),
                CXGate((c, t)),
                U3Gate(t, self.theta / 2, self.phi, 0),
            ])

    def matrix(self):
        t, p, l = self.params
        a = np.exp(0.5j * (p + l)) * np.cos(t * 0.5)
        b = np.exp(0.5j * (p - l)) * np.sin(t * 0.5)
        return np.array([[1, 0, 0, 0], [0, a.conjugate(), 0, -b.conjugate()], [0, 0, 1, 0], [0, a, 0, b]], dtype=complex)


class ToffoliGate(Gate):
    """Toffoli (CCX) gate"""
    lowername = "ccx"

    @property
    def n_qargs(self):
        return 3

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        c1, c2, t = self.targets
        return [
            HGate(t),
            CCZGate((c1, c2, t)),
            HGate(t),
        ]

    def matrix(self):
        return np.array([
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 0, 1],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 1, 0, 0, 0, 0]], dtype=complex)


class CCZGate(Gate):
    """2-Controlled Z gate"""
    lowername = "ccz"

    @property
    def n_qargs(self):
        return 3

    def fallback(self, n_qubits):
        c1, c2, t = self.targets
        return [
            CXGate((c2, t)),
            TDagGate(t),
            CXGate((c1, t)),
            TGate(t),
            CXGate((c2, t)),
            TDagGate(t),
            CXGate((c1, t)),
            TGate(c2),
            TGate(t),
            CXGate((c1, c2)),
            TGate(c1),
            TDagGate(c2),
            CXGate((c1, c2)),
        ]

    def dagger(self):
        return self

    def matrix(self):
        return np.array([
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, -1]], dtype=complex)


class CSwapGate(Gate):
    """Controlled SWAP gate"""
    lowername = "cswap"

    @property
    def n_qargs(self):
        return 3

    def dagger(self):
        return self

    def fallback(self, n_qubits):
        # TODO: test
        c, t1, t2 = self.targets
        return [CXGate((t2, t1)),
                ToffoliGate((c, t1, t2)),
                CXGate((t2, t1))]

    def matrix(self):
        return np.array([
            [1, 0, 0, 0, 0, 0, 0, 0],
            [0, 1, 0, 0, 0, 0, 0, 0],
            [0, 0, 1, 0, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 1, 0, 0],
            [0, 0, 0, 0, 1, 0, 0, 0],
            [0, 0, 0, 1, 0, 0, 0, 0],
            [0, 0, 0, 0, 0, 0, 1, 0],
            [0, 0, 0, 0, 0, 0, 0, 1]], dtype=complex)


class Measurement(Operation):
    """Measurement operation"""
    lowername = "measure"

    # Ad-hoc copy and paste programming.
    def target_iter(self, n_qubits):
        """The generator which yields the target qubits."""
        return slicing(self.targets, n_qubits)


class Reset(Operation):
    """Reset operation"""
    lowername = "reset"

    # Ad-hoc copy and paste programming.
    def target_iter(self, n_qubits):
        """The generator which yields the target qubits."""
        return slicing(self.targets, n_qubits)


def slicing_singlevalue(arg, length):
    """Internally used."""
    if isinstance(arg, slice):
        start, stop, step = arg.indices(length)
        i = start
        if step > 0:
            while i < stop:
                yield i
                i += step
        else:
            while i > stop:
                yield i
                i += step
    else:
        try:
            i = arg.__index__()
        except AttributeError:
            raise TypeError("indices must be integers or slices, not " + arg.__class__.__name__)
        if i < 0:
            i += length
        yield i


def slicing(args, length):
    """Internally used."""
    if isinstance(args, tuple):
        for arg in args:
            yield from slicing_singlevalue(arg, length)
    else:
        yield from slicing_singlevalue(args, length)


def qubit_pairs(args, length):
    """Internally used."""
    if not isinstance(args, tuple):
        raise ValueError("Control and target qubits pair(s) are required.")
    if len(args) != 2:
        raise ValueError("Control and target qubits pair(s) are required.")
    controls = list(slicing(args[0], length))
    targets = list(slicing(args[1], length))
    if len(controls) != len(targets):
        raise ValueError("The number of control qubits and target qubits are must be same.")
    for c, z in zip(controls, targets):
        if c == z:
            raise ValueError("Control qubit and target qubit are must be different.")
    return zip(controls, targets)


def get_maximum_index(indices):
    """Internally used."""
    def _maximum_idx_single(idx):
        if isinstance(idx, slice):
            start = -1
            stop = 0
            if idx.start is not None:
                start = idx.start.__index__()
            if idx.stop is not None:
                stop = idx.stop.__index__()
            return max(start, stop - 1)
        else:
            return idx.__index__()
    if isinstance(indices, tuple):
        return max((_maximum_idx_single(i) for i in indices), default=-1)
    else:
        return _maximum_idx_single(indices)


def find_n_qubits(gates):
    """Find n_qubits from gates"""
    return max((get_maximum_index(g.targets) for g in gates), default=-1) + 1
